/*
 * DMA support for the JZ4730.
 *
 * Copyright (C) 2021, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/dma-jz4730.h>
#include <l4/devices/hw_mmio_register_block.h>

#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/util/util.h>

#include <pthread.h>
#include <pthread-l4.h>

#include <stdio.h>



enum Global_regs
{
  Dma_irq_pending             = 0xf8, // IRQP
  Dma_control                 = 0xfc, // DMAC
};

enum Channel_regs
{
  Dma_source                  = 0x00, // DSA
  Dma_destination             = 0x04, // DDA
  Dma_transfer_count          = 0x08, // DTC
  Dma_request_source          = 0x0c, // DRT
  Dma_control_status          = 0x10, // DCS
};

enum Dma_irq_pending_shifts : unsigned
{
  Dma_irq_pending_ch0         = 15,
  Dma_irq_pending_ch1         = 14,
  Dma_irq_pending_ch2         = 13,
  Dma_irq_pending_ch3         = 12,
  Dma_irq_pending_ch4         = 11,
  Dma_irq_pending_ch5         = 10,
};

enum Dma_control_bits : unsigned
{
  Dma_control_priority_mode   = 0x100, // PM
  Dma_control_halt_occurred   = 0x008, // HLT
  Dma_control_address_error   = 0x004, // AR
  Dma_control_enable          = 0x001, // DMAE
};

enum Dma_control_priority_modes : unsigned
{
  Dma_priority_mode_01234567    = 0,
  Dma_priority_mode_02314675    = 1,
  Dma_priority_mode_20136457    = 2,
  Dma_priority_mode_round_robin = 3,
};

enum Dma_transfer_count_bits : unsigned
{
  Dma_transfer_count_mask     = 0x00ffffff,
};

enum Dma_request_source_bits : unsigned
{
  Dma_request_type_mask       = 0x0000001f,
};

enum Dma_control_status_shifts : unsigned
{
  Dma_ext_output_polarity     = 31,
  Dma_ext_output_mode_cycle   = 30,
  Dma_ext_req_detect_mode     = 28,
  Dma_ext_end_of_process_mode = 27,
  Dma_req_detect_int_length   = 16,
  Dma_source_port_width       = 14,
  Dma_dest_port_width         = 12,
  Dma_trans_unit_size         = 8,
  Dma_trans_mode              = 7,
};

enum Dma_control_status_bits : unsigned
{
  Dma_source_address_incr     = 0x00800000,
  Dma_source_address_no_incr  = 0x00000000,
  Dma_dest_address_incr       = 0x00400000,
  Dma_dest_address_no_incr    = 0x00000000,

  Dma_trans_unit_size_32_bit  = 0x00000000,
  Dma_trans_unit_size_8_bit   = 0x00000100,
  Dma_trans_unit_size_16_bit  = 0x00000200,
  Dma_trans_unit_size_16_byte = 0x00000300,
  Dma_trans_unit_size_32_byte = 0x00000400,

  Dma_address_error           = 0x00000010,
  Dma_trans_completed         = 0x00000008,
  Dma_trans_halted            = 0x00000004,
  Dma_channel_irq_enable      = 0x00000002,
  Dma_channel_enable          = 0x00000001,
};

enum Dma_port_width_values : unsigned
{
  Dma_port_width_32_bit       = 0,
  Dma_port_width_8_bit        = 1,
  Dma_port_width_16_bit       = 2,
};

enum Dma_trans_mode_values : unsigned
{
  Dma_trans_mode_single       = 0,
  Dma_trans_mode_block        = 1,
};



// Initialise a channel.

Dma_jz4730_channel::Dma_jz4730_channel(Dma_jz4730_chip *chip, uint8_t channel,
                                       l4_addr_t start, l4_cap_idx_t irq)
: _chip(chip), _channel(channel), _irq(irq)
{
  _regs = new Hw::Mmio_register_block<32>(start);
}

// Encode flags for an external transfer.

uint32_t
Dma_jz4730_channel::encode_external_transfer(enum Dma_jz4730_request_type type)
{
  int external = (type == Dma_request_external) ? 1 : 0;

  return
    ((external ? (int) _ext_output_polarity : 0) << Dma_ext_output_polarity) |
    ((external ? (int) _ext_output_mode_cycle : 0) << Dma_ext_output_mode_cycle) |
    ((external ? (int) _ext_req_detect_mode : 0) << Dma_ext_req_detect_mode) |
    ((external ? (int) _ext_end_of_process_mode : 0) << Dma_ext_end_of_process_mode);
}

// Return the closest interval length greater than or equal to the number of
// units given encoded in the request detection interval length field of the
// control/status register.

uint32_t
Dma_jz4730_channel::encode_req_detect_int_length(uint8_t units)
{
  static uint8_t lengths[] = {0, 2, 4, 8, 12, 16, 20, 24, 28, 32, 48, 60, 64, 124, 128, 200};
  int i;

  if (!units)
    return 0;

  for (i = 0; i <= 15; i++)
  {
     if (lengths[i] >= units)
       break;
  }

  return i << Dma_req_detect_int_length;
}

// Encode the appropriate source port width.

uint32_t
Dma_jz4730_channel::encode_source_port_width(uint8_t width)
{
  switch (width)
  {
    case 1:
    return Dma_port_width_8_bit << Dma_source_port_width;

    default:
    return Dma_port_width_32_bit << Dma_source_port_width;
  }
}

// Encode the appropriate destination port width for the given request type.

uint32_t
Dma_jz4730_channel::encode_destination_port_width(uint8_t width)
{
  switch (width)
  {
    case 1:
    return Dma_port_width_8_bit << Dma_dest_port_width;

    default:
    return Dma_port_width_32_bit << Dma_dest_port_width;
  }
}

// Encode the transfer unit size.

uint32_t
Dma_jz4730_channel::encode_transfer_unit_size(uint8_t size)
{
  switch (size)
  {
    case 1:
    return Dma_trans_unit_size_8_bit;

    case 2:
    return Dma_trans_unit_size_16_bit;

    case 16:
    return Dma_trans_unit_size_16_byte;

    case 32:
    return Dma_trans_unit_size_32_byte;

    default:
    return Dma_trans_unit_size_32_bit;
  }
}

// Transfer data between memory locations.

unsigned int
Dma_jz4730_channel::transfer(uint32_t source, uint32_t destination,
                             unsigned int count,
                             bool source_increment, bool destination_increment,
                             uint8_t source_width, uint8_t destination_width,
                             uint8_t transfer_unit_size,
                             enum Dma_jz4730_request_type type)
{
  // Ensure an absence of address error and halt conditions globally and in this channel.

  if (error() || halted())
    return 0;

  // Ensure an absence of transaction completed and zero transfer count for this channel.

  if (completed() || _regs[Dma_transfer_count])
    return 0;

  // Disable the channel.

  _regs[Dma_control_status] = _regs[Dma_control_status] & ~Dma_channel_enable;

  // Set addresses.

  _regs[Dma_source] = source;
  _regs[Dma_destination] = destination;

  // Set transfer count to the number of units.

  unsigned int units = count < Dma_transfer_count_mask ? count : Dma_transfer_count_mask;

  _regs[Dma_transfer_count] = units;

  // Set auto-request for memory-to-memory transfers. Otherwise, set the
  // indicated request type.

  _regs[Dma_request_source] = type;

  // Set control/status fields.
  // Enable the channel (and peripheral).

  /* NOTE: To be considered...
           * request detection interval length (currently left as 0)
           * increments and port widths for external transfers
           * port width overriding (for AIC...)
           * transfer mode (currently left as single)
  */

  _regs[Dma_control_status] = encode_external_transfer(type) |
                              (source_increment ? Dma_source_address_incr : Dma_source_address_no_incr) |
                              (destination_increment ? Dma_dest_address_incr : Dma_dest_address_no_incr) |
                              encode_source_port_width(source_width) |
                              encode_destination_port_width(destination_width) |
                              encode_transfer_unit_size(transfer_unit_size) |
                              (Dma_trans_mode_single << Dma_trans_mode) |
                              Dma_channel_irq_enable |
                              Dma_channel_enable;

  // Return the number of units to transfer.

  return units;
}

unsigned int
Dma_jz4730_channel::wait()
{
  // An interrupt will occur upon completion, the completion flag will be set
  // and the transfer count will be zero.

  unsigned int remaining = 0;

  do
  {
    if (!wait_for_irq(1000000))
      printf("status = %x\n", (uint32_t) _regs[Dma_control_status]);

    // Clearing the completion flag will clear the interrupt condition.
    // Any remaining units must be read before clearing the condition.

    else
    {
      remaining = _regs[Dma_transfer_count];
      ack_irq();
      break;
    }
  }
  while (!error() && !halted() && !completed());

  // Reset the channel status.

  _regs[Dma_control_status] = _regs[Dma_control_status] & ~(Dma_channel_enable |
                                  Dma_trans_completed | Dma_address_error |
                                  Dma_trans_halted);
  _regs[Dma_transfer_count] = 0;

  // Return the number of remaining units.

  return remaining;
}

// Wait indefinitely for an interrupt request, returning true if one was delivered.

bool
Dma_jz4730_channel::wait_for_irq()
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, pthread_l4_cap(pthread_self()), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, L4_IPC_NEVER)) && _chip->have_interrupt(_channel);
}

// Wait up to the given timeout (in microseconds) for an interrupt request,
// returning true if one was delivered.

bool
Dma_jz4730_channel::wait_for_irq(unsigned int timeout)
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, pthread_l4_cap(pthread_self()), 0)))
    return false;

  return !l4_error(l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4util_micros2l4to(timeout)))) && _chip->have_interrupt(_channel);
}

// Acknowledge an interrupt condition.

void
Dma_jz4730_channel::ack_irq()
{
  _regs[Dma_control_status] = _regs[Dma_control_status] & ~Dma_trans_completed;
}

// Return whether a transfer has completed.

bool
Dma_jz4730_channel::completed()
{
  return _regs[Dma_control_status] & Dma_trans_completed ? true : false;
}

// Return whether an address error condition has arisen.

bool
Dma_jz4730_channel::error()
{
  return _regs[Dma_control_status] & Dma_address_error ? true : false;
}

// Return whether a transfer has halted.

bool
Dma_jz4730_channel::halted()
{
  return _regs[Dma_control_status] & Dma_trans_halted ? true : false;
}



// Initialise the I2C controller.

Dma_jz4730_chip::Dma_jz4730_chip(l4_addr_t start, l4_addr_t end,
                                 Cpm_jz4730_chip *cpm)
: _start(start), _end(end), _cpm(cpm)
{
  _regs = new Hw::Mmio_register_block<32>(start);
}

// Enable the peripheral.

void
Dma_jz4730_chip::enable()
{
  // Make sure that the DMA clock is available.

  _cpm->start_clock(Clock_dma);

  // Enable the channel.
  // NOTE: No configuration is done for channel priority mode.

  _regs[Dma_control] = Dma_control_enable;
  while (!(_regs[Dma_control] & Dma_control_enable));
}

// Disable the channel.

void
Dma_jz4730_chip::disable()
{
  _regs[Dma_control] = 0;
  while (_regs[Dma_control] & Dma_control_enable);
}

// Obtain a channel object.

Dma_jz4730_channel *
Dma_jz4730_chip::get_channel(uint8_t channel, l4_cap_idx_t irq)
{
  if (channel < 6)
    return new Dma_jz4730_channel(this, channel, _start + 0x20 * channel, irq);
  else
    throw -L4_EINVAL;
}

// Return whether an interrupt is pending on the given channel.

bool
Dma_jz4730_chip::have_interrupt(uint8_t channel)
{
  return _regs[Dma_irq_pending] & (1 << (Dma_irq_pending_ch0 - channel)) ? true : false;
}



// C language interface functions.

void *jz4730_dma_init(l4_addr_t start, l4_addr_t end, void *cpm)
{
  return (void *) new Dma_jz4730_chip(start, end, static_cast<Cpm_jz4730_chip *>(cpm));
}

void jz4730_dma_disable(void *dma_chip)
{
  static_cast<Dma_jz4730_chip *>(dma_chip)->disable();
}

void jz4730_dma_enable(void *dma_chip)
{
  static_cast<Dma_jz4730_chip *>(dma_chip)->enable();
}

void *jz4730_dma_get_channel(void *dma, uint8_t channel, l4_cap_idx_t irq)
{
  return static_cast<Dma_jz4730_chip *>(dma)->get_channel(channel, irq);
}

void jz4730_dma_set_output_polarity(void *dma_channel, enum Dma_jz4730_ext_level polarity)
{
  static_cast<Dma_jz4730_channel *>(dma_channel)->set_output_polarity(polarity);
}

void jz4730_dma_set_end_of_process_mode(void *dma_channel, enum Dma_jz4730_ext_level mode)
{
  static_cast<Dma_jz4730_channel *>(dma_channel)->set_end_of_process_mode(mode);
}

void jz4730_dma_set_output_mode_cycle(void *dma_channel, enum Dma_jz4730_ext_output_mode_cycle cycle)
{
  static_cast<Dma_jz4730_channel *>(dma_channel)->set_output_mode_cycle(cycle);
}

void jz4730_dma_set_req_detect_mode(void *dma_channel, enum Dma_jz4730_ext_req_detect_mode mode)
{
  static_cast<Dma_jz4730_channel *>(dma_channel)->set_req_detect_mode(mode);
}

unsigned int jz4730_dma_transfer(void *dma_channel,
                                 uint32_t source, uint32_t destination,
                                 unsigned int count,
                                 int source_increment, int destination_increment,
                                 uint8_t source_width, uint8_t destination_width,
                                 uint8_t transfer_unit_size,
                                 enum Dma_jz4730_request_type type)
{
  return static_cast<Dma_jz4730_channel *>(dma_channel)->transfer(source,
    destination, count, source_increment, destination_increment, source_width,
    destination_width, transfer_unit_size, type);
}

unsigned int jz4730_dma_wait(void *dma_channel)
{
  return static_cast<Dma_jz4730_channel *>(dma_channel)->wait();
}
