/*
 * DMA-related memory allocation utility functions.
 *
 * Copyright (C) 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/dataspace.h>
#include <l4/re/c/dma_space.h>
#include <l4/re/c/mem_alloc.h>
#include <l4/re/env.h>
#include <l4/re/protocols.h>
#include <l4/sys/factory.h>
#include <l4/sys/types.h>
#include <l4/vbus/vbus.h>

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include "dma.h"
#include "memory.h"



// Obtain a DMA space to which memory regions may be associated.

long get_dma_space(l4_cap_idx_t *dma)
{
  l4_cap_idx_t vbus = l4re_env_get_cap("vbus");

  if (l4_is_invalid_cap(vbus))
    return -L4_ENOENT;

  // Obtain capabilities for the DMA region and the vbus.

  *dma = ipc_cap_alloc();

  if (l4_is_invalid_cap(*dma))
    return -L4_ENOENT;

  // Create the DMA space.

  if (l4_error(l4_factory_create(l4re_env()->mem_alloc, L4RE_PROTO_DMA_SPACE,
                                 *dma)))
    return -L4_ENOMEM;

  // Find the DMA domain and assign the DMA space.

  l4vbus_device_handle_t device = L4VBUS_NULL;
  l4vbus_resource_t dma_resource;

  if (!find_resource(&device, &dma_resource, L4VBUS_RESOURCE_DMA_DOMAIN))
    return -L4_ENOENT;

  if (l4vbus_assign_dma_domain(vbus, dma_resource.start,
                               L4VBUS_DMAD_BIND | L4VBUS_DMAD_L4RE_DMA_SPACE,
                               *dma))
    return -L4_ENOMEM;

  return L4_EOK;
}

// Obtain a memory region associated with the given DMA space.

long get_dma_region_for_space(unsigned long size, int align, l4_cap_idx_t dma,
                              struct dma_region *region)
{
  // Allocate memory at the given alignment.

  const l4_size_t alloc_flags = L4RE_MA_CONTINUOUS | L4RE_MA_PINNED;
  const l4re_rm_flags_t attach_flags = L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RW;

  region->align = 0;
  region->paddr = 0;
  region->size = 0;
  region->vaddr = 0;

  // Map the allocated memory, obtaining a virtual address.

  if (ipc_new_dataspace(size, alloc_flags, align, &region->mem))
    return -L4_ENOMEM;

  if (ipc_attach_dataspace_align(region->mem, size, attach_flags, align,
                                 (void **) &region->vaddr))
    return -L4_ENOMEM;

  // Obtain a physical address.

  l4_size_t size_out = size;

  if (l4re_dma_space_map(dma, region->mem | L4_CAP_FPAGE_RW, 0, &size_out, 0,
                         L4RE_DMA_SPACE_TO_DEVICE, &region->paddr))
    return -L4_ENOMEM;

  // Test the mapped region size.

  if (size_out < size)
    return -L4_ENOMEM;

  region->align = align;
  region->size = size_out;

  return L4_EOK;
}

// Obtain a memory region associated with its own DMA space.

long get_dma_region(unsigned long size, int align, struct dma_region *region)
{
  l4_cap_idx_t dma;
  long err = get_dma_space(&dma);

  if (err)
    return err;

  return get_dma_region_for_space(size, align, dma, region);
}
